const _ = require("lodash");

const { shortenName } = require("./Utils");
const Chance = require("chance");

module.exports = function applyNamingRules(tableMap, opts) {
  const characterPool =
    "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";

  opts = opts || {};
  var log = opts.log || function () {};
  const maxLength = opts.maxNameLength || 25;
  var objectNamingMap = {};
  var modelNamingMap = {};

  _.each(tableMap, (modelDef, modelName) => {
    if (modelDef.type === "view") return;

    applyModelNamingConvention(modelName);
    applyTableNamingConvention(modelName);
    modelDef.modelAttributes = {};
    modelDef.columnNames = {};
    _.each(modelDef.attributes, (attrDef, attrName) => {
      applyAttributeNamingConvention(modelDef, attrName);
      if (attrDef.isForeignKey) {
        applyForeignKeyNamingConvention(modelDef, attrName);
      }
    });

    _.each(modelDef.associations, (assoc) => {
      assoc.name = _.camelCase(assoc.name);
    });

    _.each(modelDef.indexes, (indexDef) => {
      applyIndexNamingConvention(modelDef, indexDef);
    });
  });

  return tableMap;

  function applyModelNamingConvention(name) {
    var formattedName = _.camelCase(name);

    tableMap[name].modelName = formattedName;

    if (doesModelNameExist(formattedName)) {
      throw new Error("Invalid model name: " + formattedName);
    } else {
      modelNamingMap[formattedName] = true;
    }
  }

  function applyTableNamingConvention(name) {
    var formattedName = _.camelCase(name);

    if (formattedName.length > maxLength) {
      formattedName = shortenName(name);
    }

    var sequenceName = formattedName + "_seq";

    var renamingTries = 99;

    let chance = new Chance(name);

    while (
      doesObjectNameExist(formattedName) ||
      doesObjectNameExist(sequenceName)
    ) {
      if (renamingTries < 0) {
        log({
          objectNamingMap,
        });
        throw new Error("Invalid table name: " + formattedName);
      }

      var postFix = chance.string({ length: 4, pool: characterPool });

      formattedName =
        formattedName.substr(0, formattedName.length - 4) + postFix;
      sequenceName = formattedName + "_seq";

      renamingTries--;
    }

    tableMap[name].tableName = formattedName;
    objectNamingMap[formattedName] = true;
    objectNamingMap[sequenceName] = true;
  }

  function applyAttributeNamingConvention(modelDef, name) {
    var formattedName = _.camelCase(name);

    modelDef.attributes[name].name = formattedName;

    var columnName = formattedName;

    if (columnName.length > maxLength) {
      columnName = shortenName(name);
    }

    modelDef.attributes[name].columnName = columnName;

    if (doesAttributeNameExist(modelDef, formattedName)) {
      throw new Error(
        `Invalid attribute name (collision): ${modelDef.modelName}.${formattedName}`
      );
    } else if (doesColumnNameExist(modelDef, columnName)) {
      throw new Error("Invalid column name: " + formattedName);
    } else {
      changeKeyName(modelDef.modelAttributes, name, formattedName);
      modelDef.columnNames[columnName] = true;
      modelDef.modelAttributes[formattedName] = true;
    }
  }

  function applyForeignKeyNamingConvention(modelDef, name) {
    var attrDef = modelDef.attributes[name];
    var columnName = name;

    if (_.endsWith(name, "_id")) {
      columnName = name.substr(0, name.length - 3);
    } else if (_.endsWith(name, "_code")) {
      columnName = name.substr(0, name.length - 4);
    }
    columnName = _.camelCase(columnName);

    var tableName = modelDef.tableName;

    var fkName = tableName + "_" + columnName;

    if (fkName.length > maxLength) {
      var targetLength = maxLength - tableName.length;
      var minTargetLength = Math.floor(maxLength / 2);
      if (targetLength < minTargetLength) targetLength = minTargetLength;

      columnName = shortenName(columnName, targetLength);

      if (columnName.length + tableName.length > minTargetLength * 2) {
        tableName = shortenName(tableName, targetLength);
      }

      fkName = tableName + "_" + columnName;
    }
    fkName += "_fk";

    if (doesObjectNameExist(fkName)) {
      let renamingTries = 0;
      let baseFkName = tableName + "_" + columnName;
      baseFkName = baseFkName.substr(0, baseFkName.length - 4);
      let chance = new Chance(fkName);
      while (doesObjectNameExist(fkName) && renamingTries++ < 99) {
        var postFix = chance.string({ length: 4, pool: characterPool });

        fkName = tableName + "_" + columnName + postFix + "_fk";
      }

      if (renamingTries > 99) {
        throw new Error("Foreign key name collision: " + fkName);
      } else {
        objectNamingMap[fkName] = true;
        attrDef.fkName = fkName;
      }
    } else if (fkName.length > maxLength + 5) {
      throw new Error(
        `Invalid foreign key name length ${fkName.length}: ` + fkName
      );
    } else {
      objectNamingMap[fkName] = true;
      attrDef.fkName = fkName;
    }
  }

  function applyIndexNamingConvention(modelDef, indexDef) {
    let columns = indexDef.attributes.map((attrName) => {
      return formatColumnName(attrName);
    });

    let tableName = modelDef.tableName;
    let columnNames = columns.join("_");
    let idxName = tableName + "_" + columnNames;
    let targetNameParts = 1 + columns.length;

    if (idxName.length > maxLength) {
      let minTargetLength = Math.floor(maxLength / targetNameParts);

      if (minTargetLength > tableName.length) {
        minTargetLength = Math.floor(
          (maxLength - tableName.length) / (targetNameParts - 1)
        );
      }

      columnNames = columns
        .map((columnName) => shortenName(columnName, minTargetLength - 1))
        .join("_");

      if (columnNames.length + tableName.length > maxLength) {
        tableName = shortenName(tableName, minTargetLength);
      }

      idxName = tableName + "_" + columnNames;
    }
    idxName += "_ix";

    if (doesObjectNameExist(idxName)) {
      let renamingTries = 0;
      let baseFkName = tableName + "_" + columnNames;
      baseFkName = baseFkName.substr(0, baseFkName.length - 4);
      let chance = new Chance(idxName);
      while (doesObjectNameExist(idxName) && renamingTries++ < 99) {
        var postFix = chance.string({ length: 4, pool: characterPool });

        idxName = baseFkName + postFix + "_ix";
      }

      if (renamingTries > 99) {
        throw new Error("Index name collision: " + idxName);
      }
    } 
    
    if (idxName.length > maxLength + 5) {
      throw new Error(
        `Invalid index name length ${idxName.length}: ` + idxName
      );
    }

    objectNamingMap[idxName] = true;
    indexDef.name = idxName;
    indexDef.columns = indexDef.attributes.map((attrName) => {
      return modelDef.attributes[attrName].columnName;
    });
  }

  function formatColumnName(name) {
    let columnName = name;
    if (_.endsWith(name, "_id")) {
      columnName = name.substr(0, name.length - 3);
    } else if (_.endsWith(name, "_code")) {
      columnName = name.substr(0, name.length - 4);
    }
    columnName = _.camelCase(columnName);

    return columnName;
  }

  function doesObjectNameExist(name) {
    return objectNamingMap[name];
  }

  function doesModelNameExist(name) {
    return modelNamingMap[name];
  }

  function doesAttributeNameExist(modelDef, name) {
    return modelDef.modelAttributes[name];
  }

  function doesColumnNameExist(modelDef, name) {
    return modelDef.columnNames[name];
  }
};

function changeKeyName(obj, oldName, newName) {
  if (obj[newName] != null) {
    throw new Error(
      "Unable to change key name " +
        oldName +
        " to " +
        newName +
        " because new key name is not null"
    );
  }

  obj[newName] = oldName;
  delete obj[oldName];
  return obj;
}
