const _ = require("lodash");
const pluralize = require("pluralize");
const log = require("../utils/log");
module.exports = function convertRowsToTableMap(
  attributes,
  entities,
  relations,
  keyAttributes,
  comments,
  opts
) {
  opts = opts || {};
  var log = opts.log || function () {};

  var tableMap = {};

  const commentsByObjId = comments.reduce((acc, comment) => {
    if (!comment.ObjectID) return acc;
    acc[comment.ObjectID] = comment.Text;
    return acc;
  }, {});

  const getInitialTableStructure = function (entityName) {
    let entity = _.find(entities, { Entity_Name: entityName });
    let comment;
    if (entity) {
      comment = commentsByObjId[entity.ObjectID];
    }

    return {
      tableName: entityName,
      attributes: {},
      associations: [],
      indexes: [],
      comment,
    };
  };

  attributes.forEach(function (attr) {
    if (typeof attr.Entity_Name === "undefined") {
      log("Undefined Entity for Attribute");
      log(attr);
      return;
    }

    var table =
      tableMap[attr.Entity_Name] || getInitialTableStructure(attr.Entity_Name);

    var tabAttr = {
      name: attr.Attribute_Name,
      dataType: attr.Logical_Type_Name,
      isPrimaryKey: attr.PK_Flag === "P",
      isForeignKey: attr.FK_Flag === "F",
      size: attr.T_Size,
      precision: attr.T_Precision,
      scale: attr.T_Scale,
      comment: commentsByObjId[attr.ObjectID],
      nullable: attr.Mandatory !== "Y",
      defaultValue: attr.Default_Value,
    };

    if (
      tabAttr.isPrimaryKey &&
      tabAttr.dataType === "NUMERIC" &&
      !tabAttr.isForeignKey
    )
      tabAttr.autoIncrement = true;
    if (tabAttr.isForeignKey) {
      var fkRelation = _.find(relations, { ObjectID: attr.Relation_ID });

      log("FK Relation for " + attr.Attribute_Name);
      log(fkRelation);

      var curDef = {},
        refDef = {};
      curDef.isPlural = false; //we the current Model def is not plural because it has the foreign key field
      //Source Entity Has The Foriegn Key
      if (fkRelation.Source_Entity_Name === attr.Entity_Name) {
        refDef.entityID = fkRelation.Target_ID;
        curDef.entityID = fkRelation.Source_ID;
        refDef.entityName = fkRelation.Target_Entity_Name;
        curDef.entityName = fkRelation.Source_Entity_Name;
        refDef.isPlural = fkRelation.TargetTo_Source_Cardinality != "1";
        curDef.identifying = false; //identifying must be false because it's the source

        //Target Entity Has The Foreign Key
      } else {
        curDef.entityID = fkRelation.Target_ID;
        refDef.entityID = fkRelation.Source_ID;
        curDef.entityName = fkRelation.Target_Entity_Name;
        refDef.entityName = fkRelation.Source_Entity_Name;
        refDef.isPlural = fkRelation.SourceTo_Target_Cardinality != "1";
        curDef.identifying = fkRelation.Identifying == "Y";
      }

      //key definitions (aka the primary key info used for the relationship)
      //maybe different from relationship info because of subtypes
      curDef.keyDef = getKeyDef(curDef.entityID);
      refDef.keyDef = getKeyDef(refDef.entityID);

      tabAttr.referenceName = refDef.keyDef.Entity_Name;
      tabAttr.referenceKey = refDef.keyDef.Attribute_Name;
      tabAttr.referenceIdentifying = curDef.identifying;
      tabAttr.referenceIsPlural = refDef.isPlural;

      var refModel = tableMap[refDef.entityName];

      if (typeof refModel === "undefined")
        refModel = getInitialTableStructure(refDef.entityName);

      var refAssociation = {
        isPlural: refDef.isPlural,
        name: curDef.entityName,
        referenceName: curDef.keyDef.Entity_Name,
        referenceKey: attr.Attribute_Name,
        comment: commentsByObjId[attr.Relation_ID],
      };

      if (refDef.keyDef.isSubType === true) {
        refAssociation.name = refDef.entityName + curDef.entityName;
        log("SubType Assoc Name: " + refAssociation.name);
      }

      if (refDef.isPlural) {
        refAssociation.name = pluralize.plural(refAssociation.name);
      }

      // TODO: we could also obtain this info more reliably from the ODM XML file.
      if (_.endsWith(fkRelation.Relation_Name, "_cascade")) {
        tabAttr.referenceShouldCascade = true;
      }

      if (_.endsWith(fkRelation.Relation_Name, "_restrict")) {
        tabAttr.referenceShouldRestrict = true;
      }

      if (_.endsWith(fkRelation.Relation_Name, "_noaction")) {
        tabAttr.referenceShouldNoAction = true;
      }

      if (_.endsWith(fkRelation.Relation_Name, "_setdefault")) {
        tabAttr.referenceShouldSetDefault = true;
      }

      refModel.associations.push(refAssociation);
      tableMap[refDef.entityName] = refModel;
    }
    if (tabAttr.isPrimaryKey) {
      table.primaryKeyField = tabAttr.name;
    }
    table.attributes[attr.Attribute_Name] = tabAttr;
    tableMap[attr.Entity_Name] = table;
  });

  function getKeyDef(entityID) {
    log("Getting KeyDef by Entity ID: " + entityID);
    var keyDef = _.find(keyAttributes, { Entity_ID: entityID });

    log("KeyDef");
    log(keyDef);

    if (typeof keyDef === "undefined") {
      var referenceEntity = _.find(entities, { ObjectID: entityID });
      keyDef = _.find(keyAttributes, {
        Entity_ID: referenceEntity.SuperTypeEntity_ID,
      });
      keyDef = _.cloneDeep(keyDef);
      keyDef.isSubType = true;
      log(
        "Super Type Key Def: " +
          referenceEntity.Entity_Name +
          " -> " +
          keyDef.Entity_Name
      );
    }
    return keyDef;
  }

  entities.forEach(function (ent) {
    if (
      ent.SuperTypeEntity_ID != "" &&
      ent.Engineering_Strategy === "Single Table"
    ) {
      var superType = _.find(entities, { ObjectID: ent.SuperTypeEntity_ID });
      var superTable = tableMap[superType.Entity_Name];
      var subTable = tableMap[ent.Entity_Name];
      // log("Entity: " + ent.Entity_Name);

      // log("superTable");
      // log(superTable);

      // log("subTable");
      // log(subTable);

      if (typeof subTable !== "undefined") {
        superTable.attributes = _.assign(
          superTable.attributes,
          subTable.attributes
        );
        if (subTable.associations)
          superTable.associations = superTable.associations.concat(
            subTable.associations
          );
        delete tableMap[ent.Entity_Name];
      }
    }
  });

  return tableMap;
};
