// src/datamodel-csv-parser/convertRowsToTableMap.js:75
// var tabAttr = {
//   name: attr.Attribute_Name,
//   dataType: attr.Logical_Type_Name,
//   isPrimaryKey: attr.PK_Flag === "P",
//   isForeignKey: attr.FK_Flag === "F",
//   size: attr.T_Size,
//   precision: attr.T_Precision,
//   scale: attr.T_Scale,
//   comment: commentsByObjId[attr.ObjectID],
//   nullable: attr.Mandatory !== "Y",
//   defaultValue: attr.Default_Value
// };
const { get, endsWith } = require("lodash");
const pluralize = require("pluralize");
const { isValidDataType } = require("../../graphql-generator/isValidDataType");

function getAttributeForViewColumn(viewName, modelName, viewColumn, tableMap) {
  const {
    name,
    relatedKeyPath,
    relatedAttributePath,
    relAttrName,
    oneToOne,
    ...rest
  } = viewColumn;

  let attr = {
    name,
    columnName: name,
    isPrimaryKey: false,
    isForeignKey: false,
    size: "",
    precision: "",
    scale: "",
    nullable: true,
    defaultValue: "",
    ...rest
  };
  if (relatedAttributePath || relatedKeyPath) {
    const relatedPath = relatedAttributePath || relatedKeyPath;
    const [entityName, attrName] = relatedPath.split(".");
    const attrPath = `${entityName}.attributes.${attrName}`;
    const attrDef = get(tableMap, attrPath);
    if (attrDef) {
      const { dataType, size, precision, scale } = attrDef;

      attr = {
        ...attr,
        dataType,
        size,
        precision,
        scale
      };

      if (relatedKeyPath) {
        if (relAttrName) {
          attr.relAttrName = relAttrName;
        } else if (endsWith(name.toLowerCase(), "id")) {
          attr.relAttrName = name.substring(0, name.length - 2);
        } else if (endsWith(name.toLowerCase(), "code")) {
          attr.relAttrName = name.substring(0, name.length - 4);
        } else {
          attr.relAttrName = tableMap[entityName].modelName;
        }

        // populate reference attributes
        attr = {
          ...attr,
          isRelatedKey: true,
          referenceIdentifying: false,
          referenceIsPlural: false,
          referenceKey: attrName,
          referenceName: entityName
        };

        addAssociation(tableMap, {
          referenceName: viewName,
          modelName,
          isPlural: !oneToOne,
          referenceKey: name,
          relatedEntity: entityName
        });
      }
    } else {
      if (relatedKeyPath) {
        throw new Error(
          `Invalid relatedKeyPath ${relatedKeyPath} (${attrPath}) for view ${viewName}, column ${name}`
        );
      } else {
        throw new Error(
          `Invalid relatedAttributePath ${relatedAttributePath} (${attrPath}) for view ${viewName}, column ${name}`
        );
      }
    }
  } else {
    if (!attr.dataType) {
      throw new Error(
        `No dataType specified for view ${viewName}, column ${name}`
      );
    } else if (!isValidDataType(attr.dataType)) {
      throw new Error(
        `Invalid dataType ${attr.dataType} specified for view ${viewName}, column ${name}`
      );
    }
  }

  return attr;
}

function addAssociation(
  tableMap,
  { referenceName, modelName, isPlural, referenceKey, relatedEntity }
) {
  let name = modelName;
  if (isPlural) {
    name = pluralize(modelName);
  }

  if (!tableMap[relatedEntity].modelAttributes[name]) {
    tableMap[relatedEntity].modelAttributes[name] = true;
  } else {
    throw new Error(
      `Associated Entity Name ${name} already exists for entity ${relatedEntity} for view ${referenceName}`
    );
  }
  tableMap[relatedEntity].associations.push({
    isPlural,
    name,
    isReadOnly: true,
    referenceName,
    referenceKey
  });
}

module.exports = {
  addAssociation,
  getAttributeForViewColumn
};
