const { each } = require("lodash");
const mustache = require("mustache");
const { getViews, setViews } = require("./getViews");
const log = require("../../utils/log");

function resolveTableAndColumnNames(tableMap) {
  const views = getViews(tableMap);

  // log("Views to format", views);

  const nameMap = getNameMapFromTableMap(tableMap);

  views.forEach(view => {
    if (!view.primaryKeyField) {
      throw new Error(
        `View does not define a primaryKeyField! While a view won't have an actual primary key it will need to designate column to serve as one for determistic sorting and paging!`
      );
    }
    view.sql = normalizeViewSql(view);
    if (view.sql.pg) {
      view.sql.pg = applyNameMapToViewSql(view.sql.pg, nameMap);
    }
    if (view.sql.oracle) {
      view.sql.oracle = applyNameMapToViewSql(view.sql.oracle, nameMap);
    }
  });
  setViews(views, tableMap);
  return tableMap;
}

// If the sql is the same for Oracle and Postgres it will
// be set directly to sql key. In which case we need to copy
// it and assign it to the dialect specific sub keys
function normalizeViewSql(view) {
  if (typeof view.sql === "string") {
    return {
      pg: view.sql,
      oracle: view.sql
    };
  }
  return view.sql;
}

function applyNameMapToViewSql(sql, nameMap) {
  let escapedQuotePreCount = (sql.match(/&quot;/g) || []).length;
  const renderedSQL = mustache.render(sql, nameMap);
  let escapedQuotePostCount = (renderedSQL.match(/&quot;/g) || []).length;

  if (escapedQuotePostCount !== escapedQuotePreCount) {
    throw new Error(
      `Error replacing view SQL with table and column names! It looks like template like {{tableName}} was used instead of {{{tableName}}}\n${sql}\n${renderedSQL}`
    );
  }

  return renderedSQL;
}

function getNameMapFromTableMap(tableMap) {
  const nameMap = { tables: {} };
  each(tableMap, (entity, entityName) => {
    if (entity.type === "view") return;

    const { tableName, modelName } = entity;

    if (modelName === "tables") {
      throw new Error(
        `Entity ${entityName} has a modelName ${modelName} which is a reserved name`
      );
    }

    nameMap.tables[modelName] = `"${tableName}"`;
    nameMap[modelName] = getAttributeNameMap(entity);
  });
  return nameMap;
}

function getAttributeNameMap(entity) {
  const attrNameMap = {};

  each(entity.attributes, (attrDef, attrName) => {
    attrNameMap[attrDef.name] = `"${attrDef.columnName}"`;
  });

  return attrNameMap;
}

module.exports = {
  resolveTableAndColumnNames,
  normalizeViewSql,
  applyNameMapToViewSql,
  getNameMapFromTableMap,
  getAttributeNameMap
};
