const path = require("path");
const _ = require("lodash");
const fse = require("fs-extra");

const parseOraCSV = require("./parseOraCSV");
const convertRowsToTableMap = require("./convertRowsToTableMap");
const applyNamingConvention = require("./applyNamingConvention");
//var log = require("../lib/log");
log = function() {};

const faker = require("faker");
const uuid = require("uuid/v4");
const chance = new require("chance")();

//testFakeData();

module.exports = generateFixtures;

function testFakeData() {
  var opts = {
    log: log
  };

  var csvPath = path.join(__dirname, "../datamodel-csv-export");
  var csvData = parseOraCSV(csvPath, opts);
  var tableMap = convertRowsToTableMap(
    csvData.attributes,
    csvData.entities,
    csvData.relations,
    csvData.keyAttributes,
    opts
  );

  generateFixtures(tableMap);
}

function generateFixtures(tableMap, seedDataPath) {
  log("Table Map");
  log(tableMap);

  fse.outputJsonSync(path.join(__dirname, "test/table-map.json"), tableMap, {
    spaces: 2
  });

  var tableNodes = categorizeTableNodes(tableMap);

  log("Table Nodes");
  log(tableNodes);

  var internalStrata = stratifyInternalNodes(tableNodes, tableMap);

  log("Internal Strata");
  log(internalStrata);

  var fakeData = generateFakeData(tableMap, tableNodes, internalStrata, seedDataPath);

  log("Fake Data");
  log(fakeData);

  log("Table Names");
  //log(Object.keys(fakeData));
  _.each(fakeData, (list, tableName) => {
    log(tableName + " : " + list.length);
  });
  //return;

  var fkAssignments = assignForeignKeys(tableMap, fakeData, seedDataPath);
  log("FK Assignments");
  log(fkAssignments);

  return {
    fkAssignments: fkAssignments,
    fakeData: fakeData
  };
}

// function getNotNullableFKs(tableMap){

//   var nonNullableFKs = [];

//   //for each foreign key
//   _.each(tableMap, (table, tableName) => {
//     _.each(table.attributes, (attr, name) => {
//       if(attr.isForeignKey && !attr.nullable){
//         nonNullableFKs.push({
//           tableName: tableName,
//           keyName: name
//         });
//       }
//     });
//   });

//   return nonNullableFKs;
// }

function assignForeignKeys(tableMap, fakeData, seedDataPath) {
  //get a pool of targets
  var targetPool = getEligibleForeignKeyTargets(fakeData);
  var fkAssignments = [];

  //for each foreign key
  _.each(tableMap, (table, tableName) => {
    if(!hasSeedData(tableName, seedDataPath)){
      _.each(table.attributes, (attr, name) => {
        if (attr.isForeignKey) {
          // initialize pools from targets for foreign key
          var fkPools = initializeFkPool(targetPool, attr);
          // if nullable randomly decide if going to assign key
          // if hasMany relationship, randomly decide if picking from fresh pool or used pools
          //  if picked from fresh pool move to used pool
          // if hasOne
          //   if identifying pick from fresh pool and move to usedOnce pool
          //   else randomly decide if picking from fresh pool or used pools
          log("Assigning FK for " + name + " in " + tableName);
          log(tableName + " : " + fakeData[tableName].length);
          log(attr.referenceName + " : " + fkPools.fresh);

          fakeData[tableName].forEach(row => {
            var fkCid;
            if (!attr.shouldCascade && !shouldAssignFK()) return;
            if (attr.referenceIsPlural) {
              //hasMany
              fkCid = randomlyPickCid(fkPools);
            } else {
              //hasOne
              if (attr.referenceIdentifying) {
                fkCid = randomlyPluckFromArray(fkPools.fresh);
                fkPools.usedOnce.push(fkCid);
              } else {
                fkCid = randomlyPickCid(fkPools);
              }
            }

            fkAssignments.push({
              fkTable: tableName,
              fkField: name,
              fkRowCid: row.cid,
              refTable: attr.referenceName,
              refField: attr.referenceKey,
              refRowCid: fkCid
            });
          });
        }
      });
    }
  });

  return fkAssignments;

  function usedAvailable(fkPools) {
    return fkPools.usedOnce.length > 0 || fkPools.used.length > 0;
  }

  function randomlyPickCid(fkPools) {
    var fkCid;
    if (
      (fkPools.fresh.length > 0 && shouldUseFresh()) ||
      !usedAvailable(fkPools)
    ) {
      fkCid = randomlyPluckFromArray(fkPools.fresh);
      fkPools.usedOnce.push(fkCid);
    } else {
      if (fkPools.usedOnce.length > 0 && fkPools.used.length > 0) {
        //we'll reuse useFresh to determine if we should use usedOnce or used
        if (shouldUseFresh()) {
          fkCid = randomlyPluckFromArray(fkPools.usedOnce);
          fkPools.used.push(fkCid);
        } else {
          fkCid = randomlyPluckFromArray(fkPools.used);
          fkPools.used.push(fkCid);
        }
      } else if (fkPools.usedOnce.length > 0) {
        fkCid = randomlyPluckFromArray(fkPools.usedOnce);
        fkPools.used.push(fkCid);
      } else {
        fkCid = randomlyPluckFromArray(fkPools.used);
        fkPools.used.push(fkCid);
      }
    }
    return fkCid;
  }
}

function randomlyPluckFromArray(list) {
  if (list.length === 0)
    throw new Error("Cannot randomly pluck from array of length 0");

  var idx = chance.integer({ min: 0, max: list.length - 1 });
  var item = _.pullAt(list, idx);
  log("Random Pluck");
  log(item);
  log(idx);
  log(list.length);
  return item[0];
}

function shouldAssignFK() {
  return chance.weighted([true, false], [7, 2]);
}

function shouldUseFresh() {
  return chance.weighted([true, false], [1, 6]);
}

function initializeFkPool(targetPool, attr) {
  return {
    fresh: clone(targetPool[attr.referenceName]),
    usedOnce: [],
    used: []
  };
}

function clone(obj) {
  return JSON.parse(JSON.stringify(obj));
}

function getEligibleForeignKeyTargets(fakeData) {
  var targets = {};

  _.each(fakeData, (data, tableName) => {
    targets[tableName] = [];

    data.forEach(item => {
      targets[tableName].push(item.cid);
    });
  });

  return targets;
}

function hasSeedData(tableName, seedDataPath){
  if(seedDataPath){
    var fullSeedDataPath = path.join(seedDataPath, applyNamingConvention(tableName) + '.json');
    return fse.existsSync(fullSeedDataPath);
  }
  return false;
}

function getSeedData(tableName, seedDataPath){

  if(seedDataPath){
    var fullSeedDataPath = path.join(seedDataPath, applyNamingConvention(tableName) + '.json');
    if(fse.existsSync(fullSeedDataPath)){
      var seedData = require(fullSeedDataPath);
      seedData.forEach((item, idx) => {
        if(item.cid == null){
          item.cid = applyNamingConvention(tableName) + "-" + idx;
        }
      });
      return seedData;
    }
  }

  return [];
}

function generateFakeData(tableMap, tableNodes, internalStrata, seedDataPath) {
  var fakeData = {};

  var rowTargetCount = 10;
  var interval = 10;

  tableNodes.leaf.forEach(leaf => {
    var pks = {};

    fakeData[leaf] = getSeedData(leaf, seedDataPath);

    if(fakeData[leaf].length === 0){
      var rowSchema = tableMap[leaf];
      // var domain;
      // var domainValues = [];

      for (var i = 0; i < rowTargetCount; i++) {
        fakeData[leaf].push(
          generateFakeRecordForAttributes(rowSchema.attributes, pks)
        );
      }
    }
  });

  for (var l = 1; l <= Object.keys(internalStrata).length; l++) {
    rowTargetCount += interval;
    var lname = "L" + l;
    if (typeof internalStrata[lname] !== "undefined") {
      _.each(internalStrata[lname], (strataNode, tableName) => {
        var pks = {};
        var rowSchema = tableMap[tableName];
        log(tableName);
        log(rowSchema);
        fakeData[tableName] = getSeedData(tableName, seedDataPath);

        if(fakeData[tableName].length === 0){
            for (var i = 0; i < rowTargetCount; i++) {
              fakeData[tableName].push(
                generateFakeRecordForAttributes(rowSchema.attributes, pks)
              );
            }
        }
      });
    }
  }

  return fakeData;
}

function generateFakeRecordForAttributes(attributes, pks) {
  let fakeRecord = {};
  _.each(attributes, (attr, name) => {
    if (!attr.isPrimaryKey && !attr.isForeignKey) {
      fakeRecord[name] = getFakeDataForType(attr);
    } else if (attr.isPrimaryKey && !attr.autoIncrement && !attr.isForeignKey) {
      var code;
      do {
        code = faker.random.alphaNumeric(7);
      } while (pks[code]);
      fakeRecord[name] = code;
      pks[code] = true;
    }
  });
  fakeRecord.cid = uuid();
  return fakeRecord;
}

function getFakeDataForType(attr, domain, domainValues) {
  var domainMap = {};
  var fakeData;

  // if(typeof domain !== 'undefined' && (
  //    typeof domainMap[domain] !== 'undefined' ||
  //    (typeof domainValues !== 'undefined' && domainValues.length > 0)
  // )){
  //     if(typeof domainValues !== 'undefined' && domainValues.length > 0){

  //     }else{

  //     }
  // }else{
  const dataTypeMap = {
    NUMERIC: attr => {
      return faker.random.number();
    },
    VARCHAR: attr => {
      return faker.random.words().substr(0, attr.size);
    },
    NVARCHAR: attr => {
      return faker.random.words().substr(0, attr.size);
    },
    CLOB: attr => {
      return faker.lorem.text();
    },
    TEXT: attr => {
      return faker.lorem.text();
    },
    INTEGER: attr => {
      return faker.random.number(1, 100000000, 0);
    },
    TINYINT: attr => {
      return faker.random.number(-1, 1, 0);
    },
    TIMESTAMP: attr => {
      return faker.date.past();
    },
    "TIMESTAMP WITH TIMEZONE": attr => {
      return faker.date.past();
    },
    JSON: attr => {
      var val = faker.commerce.productAdjective();
      return { val: val };
    },
    ORDDOC: attr => {
      var val = faker.commerce.productAdjective();
      return { val: val };
    },
    DATE: attr => {
      return faker.date.past();
    }
  };

  //if(typeof dataTypeMap[attr.dataType] == 'function'){
  //log(attr.dataType);
  if(typeof dataTypeMap[attr.dataType] != 'undefined'){
    return dataTypeMap[attr.dataType.toUpperCase()](attr);
  }
  // }

  return fakeData;
}

function categorizeTableNodes(tableMap) {
  var tableNodes = {
    leaf: [],
    internal: [],
    orphan: []
  };
  _.each(tableMap, (table, name) => {
    log(table);
    var hasFKs = false;
    var hasAssociations = table.associations.length > 0;
    _.each(table.attributes, attr => {
      if (attr.isForeignKey) {
        hasFKs = true;
      }
    });
    if (hasFKs) {
      tableNodes.internal.push(name);
    } else if (hasAssociations) {
      tableNodes.leaf.push(name);
    } else {
      tableNodes.orphan.push(name);
    }
  });

  return tableNodes;
}

function stratifyInternalNodes(tableNodes, tableMap) {
  var internalStrata = {
    L1: {},
    L2: {}
  };

  var currentLevel = 1;
  _.each(tableMap, (table, name) => {
    internalStrata.L1[name] = {
      parents: {},
      level: 1
    };
  });

  // tableNodes.leaf.forEach(function (tableName) {
  //   //var leaf = tableMap[tableName];
  //   //var parentNodeName = tableName;
  //   var parentNode = {
  //     parents: {},
  //     level: 1
  //   };

  //   internalStrata.L1[tableName] = parentNode;
  //   // stratifyAssociations(
  //   //   parentNode,
  //   //   parentNodeName,
  //   //   leaf.associations,
  //   //   currentLevel
  //   // );
  // });

  var maxLevels = tableNodes.internal.length + tableNodes.leaf.length;
  var currentLevelName = "L" + currentLevel;
  while (
    currentLevel <= maxLevels &&
    typeof internalStrata[currentLevelName] !== "undefined"
  ) {
    _.each(internalStrata[currentLevelName], function(
      parentNode,
      parentNodeName
    ) {
      if (parentNode) {
        var table = tableMap[parentNodeName];
        stratifyAssociations(
          parentNode,
          parentNodeName,
          table.associations,
          currentLevel
        );
      }
    });

    currentLevel++;
    currentLevelName = "L" + currentLevel;
  }

  if (currentLevel > maxLevels) {
    log("Exceeded Max Levels!");
  }

  return internalStrata;

  function stratifyAssociations(
    parentNode,
    parentNodeName,
    associations,
    currentLevel
  ) {
    log("Stratifying " + parentNodeName);
    associations.forEach(function(assoc) {
      try {
        var nodeName = assoc.referenceName;
        var stratifiedNode = getStratifiedNode(nodeName);
        if (typeof stratifiedNode !== "undefined") {
          if (assoc.isPlural) {
            if (stratifiedNode.level < currentLevel) {
              if (_.has(parentNode.parents, nodeName)) {
                //circular reference
                //so just make them the same level
                moveStratifiedNode(
                  stratifiedNode,
                  nodeName,
                  currentLevel,
                  parentNode,
                  parentNodeName
                );
                stratifiedNode.circular = true;
              } else {
                //no circular reference
                //just move it up above this level
                moveStratifiedNode(
                  stratifiedNode,
                  nodeName,
                  currentLevel,
                  parentNode,
                  parentNodeName
                );
              }
            } else if (stratifiedNode.level == currentLevel) {
              if (_.has(parentNode.parents, nodeName)) {
                //circular reference
                //and they're already the same level
                stratifiedNode.circular = true;
              } else {
                //no circular reference
                //just move it up above this level
                moveStratifiedNode(
                  stratifiedNode,
                  nodeName,
                  currentLevel + 1,
                  parentNode,
                  parentNodeName
                );
              }
            } else {
              if (_.has(parentNode.parents, nodeName)) {
                //circular reference
                //so just make them the same level
                moveStratifiedNode(
                  stratifiedNode,
                  nodeName,
                  currentLevel,
                  parentNode,
                  parentNodeName
                );
                stratifiedNode.circular = true;
              } else {
                //no circular reference
                //node is already above the parentNode
                //add this as a parent to existing stratified node
                stratifiedNode.parents = _.merge(
                  stratifiedNode.parents,
                  _.clone(parentNode.parents)
                );
                stratifiedNode.parents[parentNodeName] = true;
              }
            }
          } else {
            if (stratifiedNode.level < currentLevel) {
              moveStratifiedNode(
                stratifiedNode,
                nodeName,
                currentLevel,
                parentNode,
                parentNodeName
              );
            }
          }
        } else {
          if (assoc.isPlural) {
            let lname = "L" + (currentLevel + 1);

            if (typeof internalStrata[lname] === "undefined")
              internalStrata[lname] = {};

            internalStrata[lname][nodeName] = {
              parents: _.merge(_.clone(parentNode.parents), {
                [parentNodeName]: true
              }),
              level: currentLevel
            };
          } else {
            let lname = "L" + currentLevel;
            internalStrata[lname][nodeName] = {
              parents: _.merge(_.clone(parentNode.parents), {
                [parentNodeName]: true
              }),
              level: currentLevel
            };
          }
        }
      } catch (err) {
        log(err);

        log("internalStrata");
        log(internalStrata);
        log("args");
        log([parentNode, parentNodeName, associations, currentLevel]);
        log("assoc");
        log(assoc);
        throw err;
      }
    });
  }

  function getStratifiedNode(name) {
    var stratifiedNode;

    _.each(internalStrata, level => {
      if (_.has(level, name)) {
        stratifiedNode = level[name];
        return false;
      }
    });

    return stratifiedNode;
  }

  function moveStratifiedNode(
    stratifiedNode,
    nodeName,
    toLevel,
    parentNode,
    parentNodeName
  ) {
    let stratifiedNodePath = "L" + stratifiedNode.level + "." + nodeName;
    _.unset(internalStrata, stratifiedNodePath);
    stratifiedNode.level = toLevel;
    stratifiedNode.parents = _.merge(
      stratifiedNode.parents,
      _.clone(parentNode.parents)
    );
    stratifiedNode.parents[parentNodeName] = true;
    let lname = "L" + stratifiedNode.level;

    if (typeof internalStrata[lname] === "undefined")
      internalStrata[lname] = {};

    internalStrata[lname][nodeName] = stratifiedNode;
  }
}
