const _ = require("lodash");
const getCleanSchemaCommand = require("./getCleanSchemaCommand");
const getTriggerFunctions = require("./getTriggerFunctions");
const convertToKnexType = require("./convertToKnexType");
const { getViewFromModel } = require("./getViewFromModel");
module.exports = function (
  modelSpec,
  schemaName,
  commonSchemaName,
  dialect,
  opts
) {
  if (commonSchemaName && dialect !== "postgres") {
    throw new Error(
      `Dialect ${dialect} does not support common schema ${commonSchemaName}`
    );
  }

  var tableTemp = {
    schema: "schemaName",
    tableName: "tableName",
    columns: ["columnTemp"],
  };

  var columnTemp = {
    type: "string | integer | specificType",
    typeArgs: [],
    name: "",
    primaryKey: false,
    increments: false,
  };

  var associationTemp = {
    tableName: "tableName",
    columnName: "foreignKeyAttribute",
    referenceTableName: "refTableName",
    referenceColumnName: "refColumnName",
    fkName: "fkName",
    onUpdate: "",
    onDelete: "SET NULL",
  };

  let indexTemp = {
    indexName: "indexName",
    columns: ["columnNames"],
  };

  var knexSpec = {
    tables: [],
    relationships: [],
    constraints: [],
    views: [],
    cleanSchema: getCleanSchemaCommand(dialect, schemaName),
    cleanCommonSchema: getCleanSchemaCommand(dialect, commonSchemaName),
    extensions: [
      {
        cmd: "CREATE EXTENSION IF NOT EXISTS pgcrypto WITH SCHEMA ?? ",
        isCommon: true,
        bindings: [commonSchemaName],
      },
      {
        cmd: "CREATE EXTENSION IF NOT EXISTS pgcrypto WITH SCHEMA ?? ",
        bindings: [schemaName],
      },
    ],
  };

  _.each(modelSpec, (modelDef, modelName) => {
    if (modelDef.type === "view") {
      let viewSchemaName = schemaName;
      let isCommon = false;
      if (modelDef.isCommonEntity) {
        isCommon = true;
        viewSchemaName = commonSchemaName;
      }
      knexSpec.views.push(
        getViewFromModel(modelDef, viewSchemaName, dialect, isCommon)
      );
    } else {
      var tableSpec = clone(tableTemp);
      tableSpec.schema = schemaName;
      if (modelDef.isCommonEntity) {
        tableSpec.isCommon = true;
        tableSpec.schema = commonSchemaName;
      }
      tableSpec.tableName = modelDef.tableName;
      tableSpec.columns = [];
      tableSpec.indexes = [];

      _.each(modelDef.attributes, (attrDef, attrName) => {
        var colSpec = clone(columnTemp);
        colSpec = _.assign(
          colSpec,
          convertToKnexType(attrDef, attrDef.columnName, dialect)
        );
        colSpec.name = attrDef.columnName;
        colSpec.primaryKey = attrDef.isPrimaryKey;
        colSpec.increments = attrDef.autoIncrement;
        colSpec.unique = attrDef.unique;
        if (attrDef.defaultValue) {
          colSpec.defaultValue = attrDef.defaultValue;
        } else if (attrDef.defaultRawValue) {
          colSpec.defaultRawValue = attrDef.defaultRawValue;
        }
        tableSpec.columns.push(colSpec);

        if (attrDef.isForeignKey) {
          var relSpec = clone(associationTemp);
          relSpec.fkName = attrDef.fkName;
          relSpec.tableName = modelDef.tableName;
          relSpec.schema = tableSpec.schema;
          relSpec.columnName = attrDef.columnName;
          relSpec.referenceTableName =
            modelSpec[attrDef.referenceName].tableName;
          relSpec.referenceColumnName =
            modelSpec[attrDef.referenceName].attributes[
              attrDef.referenceKey
            ].columnName;
          if (!attrDef.referenceSchema) {
            if (modelSpec[attrDef.referenceName].isCommonEntity) {
              relSpec.referenceSchema = commonSchemaName;
            } else {
              relSpec.referenceSchema = schemaName;
            }
          } else {
            relSpec.referenceSchema = attrDef.referenceSchema;
          }
          if (
            relSpec.referenceSchema === relSpec.schema &&
            tableSpec.isCommon
          ) {
            relSpec.isCommon = true;
          }
          if (attrName == "code") relSpec.onUpdate = "CASCADE";
          if (!attrDef.nullable || attrDef.referenceShouldCascade === true)
            relSpec.onDelete = "CASCADE";
          if (attrDef.referenceShouldNoAction === true)
              relSpec.onDelete = "NO ACTION";
          if (attrDef.referenceShouldRestrict === true)
            relSpec.onDelete = "RESTRICT";
          if (attrDef.referenceShouldSetDefault === true)
            relSpec.onDelete = "SET DEFAULT";

          knexSpec.relationships.push(relSpec);
        }

        if (!attrDef.nullable && !attrDef.isPrimaryKey) {
          var constraintSpec = clone(colSpec);
          constraintSpec.isCommon = tableSpec.isCommon;
          constraintSpec.schema = tableSpec.schema;
          constraintSpec.tableName = modelDef.tableName;
          constraintSpec.constraintType = "NOT_NULL";
          knexSpec.constraints.push(constraintSpec);
        }
      });

      _.each(modelDef.indexes, (indexDef) => {
        let indexSpec = clone(indexTemp);
        indexSpec.indexName = indexDef.name;
        indexSpec.isCommon = tableSpec.isCommon;
        indexSpec.columns = indexDef.columns;
        indexSpec.unique = indexDef.unique;
        indexSpec.sparseIndex = indexDef.sparseIndex;
        tableSpec.indexes.push(indexSpec);
      });
      knexSpec.tables.push(tableSpec);
    }
  });

  _.each(modelSpec, (modelDef, entityName) => {
    if (modelDef.type !== "view") {
      const modelName = modelDef.modelName;
      const tableName = modelDef.tableName;
      let isCommon = false;
      let triggerFunctionSchemaName = schemaName;
      if (modelDef.isCommonEntity) {
        isCommon = true;
        triggerFunctionSchemaName = commonSchemaName;
      }

      const triggerFunction = getTriggerFunctions(
        dialect,
        triggerFunctionSchemaName,
        modelName,
        tableName
      );

      knexSpec.functions = knexSpec.functions || [];
      knexSpec.triggers = knexSpec.triggers || [];
      knexSpec.initSql = knexSpec.initSql || [];

      knexSpec.functions = knexSpec.functions.concat(
        addCommonFlag(triggerFunction.functions, isCommon)
      );
      knexSpec.triggers = knexSpec.triggers.concat(
        addCommonFlag(triggerFunction.triggers, isCommon)
      );
      knexSpec.initSql = knexSpec.initSql.concat(
        addCommonFlag(triggerFunction.initSql, isCommon)
      );
    }
  });

  return knexSpec;
};

function clone(temp) {
  return JSON.parse(JSON.stringify(temp));
}

function addCommonFlag(cmds = [], isCommon) {
  return cmds.map((cmd) => {
    cmd.isCommon = isCommon;
    return cmd;
  });
}
