const initSQLLogging = require("./initSQLLogging");
const applySchemaTablesAndFunctions = require("./applySchemaTablesAndFunctions");
const cleanModuleSchemas = require("./cleanModuleSchemas");
const applySchemaRefsAndConstraints = require("./applySchemaRefsAndConstraints");
const applySchemaTriggers = require("./applySchemaTriggers");
const applyViews = require("./applyViews");
const runInitSql = require("./runInitSql");
const createExtensions = require("./createExtensions");

async function applySchema(db, schemaSpec, initData, options = {}, appConfig) {
  let {
    log = () => {},
    cleanSchema,
    commonSchema,
    moduleSchema,
    ...rest
  } = options;

  const { cleanCommonSchema } = appConfig;

  if (cleanCommonSchema) {
    cleanSchema = true;
    commonSchema = true;
  }

  let opts = {
    ...rest,
    log,
    cleanSchema,
    commonSchema,
    moduleSchema,
  };

  log(
    `[ORADM-APPLY-SCHEMA] Applying moduleSchema: ${moduleSchema}, commonSchema: ${commonSchema}`
  );

  //init SQL Logging for logging out SQL commands for migration scripts
  initSQLLogging(db);

  if (!opts.skipCleanSchema) {
    log(`[ORADM-APPLY-SCHEMA] Cleaning Module Schemas`);
    await cleanModuleSchemas(db, schemaSpec, opts, appConfig);
  }

  if (opts.createExtensions) {
    log(`[ORADM-APPLY-SCHEMA] Creating extensions`);
    await createExtensions(db, schemaSpec, opts);
  }

  if (!opts.skipTables) {
    log(`[ORADM-APPLY-SCHEMA] Creating tables and functions`);
    //clean existing schema and create tables and functions
    await applySchemaTablesAndFunctions(db, schemaSpec, opts, appConfig);

    //if an initData function is provided call it to seed data before
    if (initData) {
      log(`[ORADM-APPLY-SCHEMA] Initializing data`);
      await initData(db, schemaSpec, opts, appConfig);
    }
  }

  if (!opts.skipConstraints) {
    log(`[ORADM-APPLY-SCHEMA] Creating foreign keys and constraints`);
    //after data has been seeded and nuids resolved apply refs and constraints
    await applySchemaRefsAndConstraints(db, schemaSpec, opts, appConfig);

    log(`[ORADM-APPLY-SCHEMA] Creating triggers`);
    //create any triggers
    await applySchemaTriggers(db, schemaSpec, opts, appConfig);
  }

  if (!opts.skipViews) {
    log(`[ORADM-APPLY-SCHEMA] Creating views`);
    //create any views -- make sure to do this after constraints
    await applyViews(db, schemaSpec, opts, appConfig);

    log(`[ORADM-APPLY-SCHEMA] Creating running final initialization SQL`);
    //run sql to initialize data values
    await runInitSql(db, schemaSpec.initSql, opts);
  }
}

module.exports = applySchema;
