const Promise = require("bluebird");

function createTables(db, tableSpecs, opts = {}) {
  const { logDebug = () => {}, commonSchema, moduleSchema } = opts;

  db.logSQL(`Creating tables`);
  return Promise.each(tableSpecs, async (tableSpec) => {
    db.logSQL(`Creating table ${tableSpec.tableName}`);

    if (tableSpec.isCommon && !commonSchema) {
      logDebug("[ORADM-TABLE] Skipping common Table " + tableSpec.tableName);
      return;
    }
    if (!tableSpec.isCommon && !moduleSchema) {
      logDebug("[ORADM-TABLE] Skipping module Table " + tableSpec.tableName);
      return;
    }
    // logDebug("[ORADM-TABLE] Creating Table " + tableSpec.tableName, tableSpec);

    let rawCommands = [];

    var qb = db.schema
      .withSchema(tableSpec.schema)
      .createTable(tableSpec.tableName, function (table) {
        var self = this;
        logDebug("[ORADM-TABLE] Creating Table " + tableSpec.tableName);
        tableSpec.columns.forEach(function (colSpec) {
          var col;

          logDebug("[ORADM-TABLE] Creating Column " + colSpec.name);
          logDebug(colSpec);
          // logDebug(table[colSpec.type].apply);
          if (colSpec.increments) {
            col = table.increments(colSpec.name, {
              primaryKey: colSpec.primaryKey,
            });
          } else {
            col = table[colSpec.type].apply(self, colSpec.typeArgs);
          }

          if (colSpec.primaryKey && !colSpec.increments) {
            col = col.primary();
          }

          if (colSpec.unique) {
            table.unique(colSpec.name);
          }
          if (colSpec.type === "boolean") {
            colSpec.defaultValue = colSpec.defaultValue || false;
          }
          if (colSpec.defaultValue || colSpec.defaultValue === false) {
            let defaultValue = colSpec.defaultValue;
            if (colSpec.type === "boolean") {
              if (
                defaultValue === "false" ||
                defaultValue === "0" ||
                defaultValue === false
              ) {
                defaultValue = false;
              } else {
                defaultValue = true;
              }
            }
            col.notNullable().defaultTo(defaultValue);
          } else if (colSpec.defaultRawValue) {
            col.notNullable().defaultTo(db.raw(colSpec.defaultRawValue));
          }
        });
        tableSpec.indexes.forEach((indexSpec) => {
          logDebug("[ORADM-TABLE] Creating Index " + indexSpec.indexName);
          logDebug(indexSpec);
          if (indexSpec.unique) {
            table.unique(indexSpec.columns, indexSpec.indexName);
          } else if (indexSpec.sparseIndex) {
            let sparseIndexCmd = `CREATE INDEX "${indexSpec.indexName}" ON "${
              tableSpec.schema
            }"."${tableSpec.tableName}" (${indexSpec.columns
              .map((c) => `"${c}"`)
              .join(", ")}) WHERE ${indexSpec.columns
              .map((c) => `"${c}" IS NOT NULL `)
              .join(" AND ")}`;
            rawCommands.push(sparseIndexCmd);
            // console.log(`Adding sparse index cmd: `, sparseIndexCmd);
          } else {
            table.index(indexSpec.columns, indexSpec.indexName);
          }
        });
      });
    await db.logSQL(qb);
    await Promise.each(rawCommands, (cmd) => {
      logDebug("[ORADM-TABLE] executing raw command: " + cmd);
      let qb = db.raw(cmd);
      return qb;
      // return db.logSQL(qb);
    });
  });
}

module.exports = createTables;
